# Common abilities
ALL = 'Appliance.System.All'
ABILITY = 'Appliance.System.Ability'
REPORT = 'Appliance.System.Report'
ONLINE = 'Appliance.System.Online'
WIFI_LIST = 'Appliance.Config.WifiList'
DEBUG = 'Appliance.System.Debug'
TRACE = 'Appliance.Config.Trace'

# Power plug/bulbs abilities
TOGGLE = 'Appliance.Control.Toggle'
TOGGLEX = 'Appliance.Control.ToggleX'
TRIGGER = 'Appliance.Control.Trigger'
TRIGGERX = 'Appliance.Control.TriggerX'
ELECTRICITY = 'Appliance.Control.Electricity'
CONSUMPTIONX = 'Appliance.Control.ConsumptionX'

# Hub
HUB_TOGGLEX = 'Appliance.Hub.ToggleX'
HUB_MTS100_TEMPERATURE = 'Appliance.Hub.Mts100.Temperature'
HUB_MTS100_MODE = 'Appliance.Hub.Mts100.Mode'
HUB_MTS100_ALL = 'Appliance.Hub.Mts100.All'

# Garage opener abilities
GARAGE_DOOR_STATE = 'Appliance.GarageDoor.State'

# Bulbs-only abilities
LIGHT = 'Appliance.Control.Light'
